import torch, os, sys, torchvision, argparse
import torchvision.transforms as tfs
from metrics import psnr, ssim, msssim
from model.models import *
import time, math
import numpy as np
from torch.backends import cudnn
from torch import optim
import torch, warnings
from torch import nn
from data_utils.RESIDE_data import *
import json
# from tensorboardX import SummaryWriter
import torchvision.utils as utils

warnings.filterwarnings('ignore')
from option import opt, model_name, log_dir
from data_utils import *
from torchvision.models import vgg16

print('log_dir :', log_dir)
print('model_name:', model_name)

models_ = {
	'IPB': IPDNet(3),
}
loaders_ = {
	'its_train': ITS_train_loader,
	'its_test': ITS_test_loader,
	'ots_train': OTS_train_loader,
	'ots_test': OTS_test_loader,
}
start_time = time.time()
T = opt.steps


def lr_schedule_cosdecay(t, T, init_lr=opt.lr):
	lr = 0.5 * (1 + math.cos(t * math.pi / T)) * init_lr
	return lr


def train(net, loader_train, loader_test, optim, criterion):
	losses = []
	start_step = 0
	max_ssim = 0
	max_psnr = 0
	ssims = []
	psnrs = []
	if opt.resume and os.path.exists(opt.model_dir):
		print(f'resume from {opt.model_dir}')
		ckp = torch.load(opt.model_dir)
		losses = ckp['losses']
		net.load_state_dict(ckp['model'])
		start_step = ckp['step']
		max_ssim = ckp['max_ssim']
		max_psnr = ckp['max_psnr']
		psnrs = ckp['psnrs']
		ssims = ckp['ssims']
		print(f'start_step:{start_step} start training ---')
	else:
		print('train from scratch *** ')
	for step in range(start_step + 1, opt.steps + 1):
		net.train()
		lr = opt.lr
		if not opt.no_lr_sche:
			lr = lr_schedule_cosdecay(step, T)
			for param_group in optim.param_groups:
				param_group["lr"] = lr
		x, y = next(iter(loader_train))
		x = x.to(opt.device)
		y = y.to(opt.device)
		out1, out2, out3 = net(x)
		Loss = nn.L1Loss(size_average=True,reduce=True)
		#L2Loss = nn.MSELoss(size_average=True,reduce=True)
		#Loss = nn.SmoothL1Loss(size_average=True,reduce=True)
		loss1 = Loss(out1, y)
		loss2 = Loss(out2, y)
		loss3 = Loss(out3, y)
		loss_L1 = loss3 + 0.6 * loss1 + 0.4 * loss2
		ssim_loss = ssim
		ssim_loss1 = 1 - ssim_loss(out1, y)

		ssim_loss2 = 1 - ssim_loss(out2, y)
		ssim_loss3 = 1 - ssim_loss(out3, y)
		loss_SSIM = ssim_loss3 + 0.4 * ssim_loss2 + 0.6 * ssim_loss1
		loss = 0.5 * loss_SSIM + loss_L1
		loss.backward()
		torch.nn.utils.clip_grad_norm(net.parameters(),max_norm=1000,norm_type=2)

		optim.step()
		optim.zero_grad()
		losses.append(loss.item())
		print(
			f'\rtrain loss : {loss.item():.5f}| step :{step}/{opt.steps}|lr :{lr :.7f} |time_used :{(time.time() - start_time) / 60 :.1f}',
			end='', flush=True)

		# with SummaryWriter(logdir=log_dir,comment=log_dir) as writer:
		#	writer.add_scalar('data/loss',loss,step)

		if step % opt.eval_step == 0:
			with torch.no_grad():
				ssim_eval, psnr_eval = test(net, loader_test, max_psnr, max_ssim, step)

			print(f'\nstep :{step} |ssim:{ssim_eval:.4f}| psnr:{psnr_eval:.4f}')

			# with SummaryWriter(logdir=log_dir,comment=log_dir) as writer:
			# 	writer.add_scalar('data/ssim',ssim_eval,step)
			# 	writer.add_scalar('data/psnr',psnr_eval,step)
			# 	writer.add_scalars('group',{
			# 		'ssim':ssim_eval,
			# 		'psnr':psnr_eval,
			# 		'loss':loss
			# 	},step)
			ssims.append(ssim_eval)
			psnrs.append(psnr_eval)
			if ssim_eval > max_ssim and psnr_eval > max_psnr:
				max_ssim = max(max_ssim, ssim_eval)
				max_psnr = max(max_psnr, psnr_eval)
				torch.save({
					'step': step,
					'max_psnr': max_psnr,
					'max_ssim': max_ssim,
					'ssims': ssims,
					'psnrs': psnrs,
					'losses': losses,
					'model': net.state_dict()
				}, opt.model_dir)
				print(f'\n model saved at step :{step}| max_psnr:{max_psnr:.4f}|max_ssim:{max_ssim:.4f}')

	np.save(f'./numpy_files/{model_name}_{opt.steps}_losses.npy', losses)
	np.save(f'./numpy_files/{model_name}_{opt.steps}_ssims.npy', ssims)
	np.save(f'./numpy_files/{model_name}_{opt.steps}_psnrs.npy', psnrs)


def test(net, loader_test, max_psnr, max_ssim, step):
	net.eval()
	torch.cuda.empty_cache()
	ssims = []
	psnrs = []
	# s=True
	for i, (inputs, targets) in enumerate(loader_test):
		inputs = inputs.to(opt.device)
		targets = targets.to(opt.device)
		x1, x2, pred = net(inputs)
		# # print(pred)
		# tfs.ToPILImage()(torch.squeeze(targets.cpu())).save('111.png')
		# vutils.save_image(targets.cpu(),'target.png')
		# vutils.save_image(pred.cpu(),'pred.png')
		ssim1 = ssim(pred, targets).item()
		psnr1 = psnr(pred, targets)
		ssims.append(ssim1)
		psnrs.append(psnr1)
	# if (psnr1>max_psnr or ssim1 > max_ssim) and s :
	#		ts=vutils.make_grid([torch.squeeze(inputs.cpu()),torch.squeeze(targets.cpu()),torch.squeeze(pred.clamp(0,1).cpu())])
	#		vutils.save_image(ts,f'samples/{model_name}/{step}_{psnr1:.4}_{ssim1:.4}.png')
	#		s=False
	return np.mean(ssims), np.mean(psnrs)


if __name__ == "__main__":
	loader_train = loaders_[opt.trainset]
	loader_test = loaders_[opt.testset]
	net = models_[opt.net]
	net = net.to(opt.device)
	epoch_size = len(loader_train)
	print("epoch_size: ", epoch_size)
	if opt.device == 'cuda':
		net = torch.nn.DataParallel(net)
		cudnn.benchmark = True

	pytorch_total_params = sum(p.numel() for p in net.parameters() if p.requires_grad)
	print("Total_params: ==> {}".format(pytorch_total_params))


	criterion = []
	criterion.append(nn.L1Loss().to(opt.device))


	optimizer = optim.Adam(params = filter(lambda x: x.requires_grad, net.parameters()), lr=opt.lr, betas=(0.9, 0.999), eps=1e-08)
	optimizer.zero_grad()
	train(net, loader_train, loader_test, optimizer, criterion)
